@@ -1,7 +1,5 @@
-try:
-    import builtins
-except ImportError:
-    import __builtin__ as builtins
+import builtins
+import collections
 import itertools
 import re
 import sys
@@ -13,17 +11,6 @@ import sphinx.util.logging
 _LOGGER = sphinx.util.logging.getLogger(__name__)
 
 
-if sys.version_info < (3,):
-    _EXCEPTIONS_MODULE = "exceptions"
-    # getattr to keep linter happy
-    _STRING_TYPES = getattr(builtins, "basestring")
-    _zip_longest = itertools.izip_longest  # pylint: disable=invalid-name,no-member
-else:
-    _EXCEPTIONS_MODULE = "builtins"
-    _STRING_TYPES = (str,)
-    _zip_longest = itertools.zip_longest  # pylint: disable=invalid-name
-
-
 def resolve_import_alias(name, import_names):
     """Resolve a name from an aliased import to its original name.
 
@@ -196,8 +183,7 @@ def get_assign_annotation(node):
     try:
         annotation_node = node.annotation
     except AttributeError:
-        # Python 2 has no support for type annotations, so use getattr
-        annotation_node = getattr(node, "type_annotation", None)
+        annotation_node = node.type_annotation
 
     if annotation_node:
         if isinstance(annotation_node, astroid.nodes.Const):
@@ -275,6 +261,42 @@ def is_decorated_with_property_setter(node):
     return False
 
 
+def is_decorated_with_overload(node):
+    """Check if the function is decorated as an overload definition.
+
+    :param node: The node to check.
+    :type node: astroid.nodes.FunctionDef
+
+    :returns: True if the function is an overload definition, False otherwise.
+    :rtype: bool
+    """
+    if not node.decorators:
+        return False
+
+    for decorator in node.decorators.nodes:
+        if not isinstance(decorator, (astroid.Name, astroid.Attribute)):
+            continue
+
+        try:
+            if _is_overload_decorator(decorator):
+                return True
+        except astroid.InferenceError:
+            pass
+
+    return False
+
+
+def _is_overload_decorator(decorator):
+    for inferred in decorator.infer():
+        if not isinstance(inferred, astroid.nodes.FunctionDef):
+            continue
+
+        if inferred.name == "overload" and inferred.root().name == "typing":
+            return True
+
+    return False
+
+
 def is_constructor(node):
     """Check if the function is a constructor.
 
@@ -300,10 +322,7 @@ def is_exception(node):
     :returns: True if the class is an exception, False otherwise.
     :rtype: bool
     """
-    if (
-        node.name in ("Exception", "BaseException")
-        and node.root().name == _EXCEPTIONS_MODULE
-    ):
+    if node.name in ("Exception", "BaseException") and node.root().name == "builtins":
         return True
 
     if not hasattr(node, "ancestors"):
@@ -360,7 +379,7 @@ def get_module_all(node):
                     continue
 
                 if isinstance(elt_name, astroid.Const) and isinstance(
-                    elt_name.value, _STRING_TYPES
+                    elt_name.value, str
                 ):
                     all_.append(elt_name.value)
 
@@ -375,7 +394,7 @@ def _is_ellipsis(node):
 
 
 def merge_annotations(annotations, comment_annotations):
-    for ann, comment_ann in _zip_longest(annotations, comment_annotations):
+    for ann, comment_ann in itertools.zip_longest(annotations, comment_annotations):
         if ann and not _is_ellipsis(ann):
             yield ann
         elif comment_ann and not _is_ellipsis(comment_ann):
@@ -384,41 +403,33 @@ def merge_annotations(annotations, comment_annotations):
             yield None
 
 
-def _format_args(args, defaults=None, annotations=None):
-    values = []
-
-    if args is None:
-        return ""
-
-    if annotations is None:
-        annotations = []
-
-    if defaults is not None:
-        default_offset = len(args) - len(defaults)
-
-    packed = _zip_longest(args, annotations)
-    for i, (arg, annotation) in enumerate(packed):
-        if isinstance(arg, astroid.Tuple):
-            values.append("({})".format(_format_args(arg.elts)))
+def _format_annotation(annotation):
+    if annotation:
+        if isinstance(annotation, astroid.Const):
+            annotation = annotation.value
         else:
-            argname = arg.name
-            default_sep = "="
-            if annotation is not None:
-                ann_str = annotation.as_string()
-                if isinstance(annotation, astroid.Const):
-                    ann_str = annotation.value
-                argname = "{}: {}".format(argname, ann_str)
-                default_sep = " = "
-            values.append(argname)
+            annotation = annotation.as_string()
 
-            if defaults is not None and i >= default_offset:
-                if defaults[i - default_offset] is not None:
-                    values[-1] += default_sep + defaults[i - default_offset].as_string()
-
-    return ", ".join(values)
+    return annotation
 
 
-def format_args(args_node):  # pylint: disable=too-many-branches,too-many-statements
+def _iter_args(args, annotations, defaults):
+    default_offset = len(args) - len(defaults)
+    packed = itertools.zip_longest(args, annotations)
+    for i, (arg, annotation) in enumerate(packed):
+        default = None
+        if defaults is not None and i >= default_offset:
+            if defaults[i - default_offset] is not None:
+                default = defaults[i - default_offset].as_string()
+
+        name = arg.name
+        if isinstance(arg, astroid.Tuple):
+            name = "({})".format(", ".join(x.name for x in arg.elts))
+
+        yield (name, _format_annotation(annotation), default)
+
+
+def _get_args_info(args_node):  # pylint: disable=too-many-branches,too-many-statements
     result = []
     positional_only_defaults = []
     positional_or_keyword_defaults = args_node.defaults
@@ -429,19 +440,11 @@ def format_args(args_node):  # pylint: disable=too-many-branches,too-many-statem
             : len(args_node.defaults) - len(args)
         ]
 
-    plain_annotations = getattr(args_node, "annotations", ()) or ()
-    func_comment_annotations = getattr(args_node.parent, "type_comment_args", ()) or ()
-    comment_annotations = getattr(args_node, "type_comment_args", []) or []
-    if hasattr(args_node, "type_comment_posonlyargs"):
-        comment_annotations = args_node.type_comment_posonlyargs + comment_annotations
-    else:
-        # astroid used to not expose type comments of positional only arguments,
-        # so pad the comments with the number of positional only arguments.
-        comment_annotations = (
-            [None] * len(getattr(args_node, "posonlyargs", ()))
-        ) + comment_annotations
-    if hasattr(args_node, "type_comment_kwonlyargs"):
-        comment_annotations += args_node.type_comment_kwonlyargs
+    plain_annotations = args_node.annotations or ()
+    func_comment_annotations = args_node.parent.type_comment_args or ()
+    comment_annotations = args_node.type_comment_posonlyargs
+    comment_annotations += args_node.type_comment_args or []
+    comment_annotations += args_node.type_comment_kwonlyargs
     annotations = list(
         merge_annotations(
             plain_annotations,
@@ -450,7 +453,7 @@ def format_args(args_node):  # pylint: disable=too-many-branches,too-many-statem
     )
     annotation_offset = 0
 
-    if getattr(args_node, "posonlyargs", None):
+    if args_node.posonlyargs:
         posonlyargs_annotations = args_node.posonlyargs_annotations
         if not any(args_node.posonlyargs_annotations):
             num_args = len(args_node.posonlyargs)
@@ -458,43 +461,39 @@ def format_args(args_node):  # pylint: disable=too-many-branches,too-many-statem
                 annotation_offset : annotation_offset + num_args
             ]
 
-        result.append(
-            _format_args(
-                args_node.posonlyargs, positional_only_defaults, posonlyargs_annotations
-            )
-        )
-        result.append("/")
+        for arg, annotation, default in _iter_args(
+            args_node.posonlyargs, posonlyargs_annotations, positional_only_defaults
+        ):
+            result.append((None, arg, annotation, default))
+
+        result.append(("/", None, None, None))
 
         if not any(args_node.posonlyargs_annotations):
             annotation_offset += num_args
 
     if args_node.args:
         num_args = len(args_node.args)
-        result.append(
-            _format_args(
-                args_node.args,
-                positional_or_keyword_defaults,
-                annotations[annotation_offset : annotation_offset + num_args],
-            )
-        )
+        for arg, annotation, default in _iter_args(
+            args_node.args,
+            annotations[annotation_offset : annotation_offset + num_args],
+            positional_or_keyword_defaults,
+        ):
+            result.append((None, arg, annotation, default))
+
         annotation_offset += num_args
 
     if args_node.vararg:
-        vararg_result = "*{}".format(args_node.vararg)
-        if getattr(args_node, "varargannotation", None):
-            vararg_result = "{}: {}".format(
-                vararg_result, args_node.varargannotation.as_string()
-            )
+        annotation = None
+        if args_node.varargannotation:
+            annotation = _format_annotation(args_node.varargannotation)
         elif len(annotations) > annotation_offset and annotations[annotation_offset]:
-            vararg_result = "{}: {}".format(
-                vararg_result, annotations[annotation_offset].as_string()
-            )
+            annotation = _format_annotation(annotations[annotation_offset])
             annotation_offset += 1
-        result.append(vararg_result)
+        result.append(("*", args_node.vararg, annotation, None))
 
-    if getattr(args_node, "kwonlyargs", None):
+    if args_node.kwonlyargs:
         if not args_node.vararg:
-            result.append("*")
+            result.append(("*", None, None, None))
 
         kwonlyargs_annotations = args_node.kwonlyargs_annotations
         if not any(args_node.kwonlyargs_annotations):
@@ -503,31 +502,55 @@ def format_args(args_node):  # pylint: disable=too-many-branches,too-many-statem
                 annotation_offset : annotation_offset + num_args
             ]
 
-        result.append(
-            _format_args(
-                args_node.kwonlyargs, args_node.kw_defaults, kwonlyargs_annotations,
-            )
-        )
+        for arg, annotation, default in _iter_args(
+            args_node.kwonlyargs,
+            kwonlyargs_annotations,
+            args_node.kw_defaults,
+        ):
+            result.append((None, arg, annotation, default))
 
         if not any(args_node.kwonlyargs_annotations):
             annotation_offset += num_args
 
     if args_node.kwarg:
-        kwarg_result = "**{}".format(args_node.kwarg)
-        if getattr(args_node, "kwargannotation", None):
-            kwarg_result = "{}: {}".format(
-                kwarg_result, args_node.kwargannotation.as_string()
-            )
+        annotation = None
+        if args_node.kwargannotation:
+            annotation = _format_annotation(args_node.kwargannotation)
         elif len(annotations) > annotation_offset and annotations[annotation_offset]:
-            kwarg_result = "{}: {}".format(
-                kwarg_result, annotations[annotation_offset].as_string()
-            )
+            annotation = _format_annotation(annotations[annotation_offset])
             annotation_offset += 1
-        result.append(kwarg_result)
+        result.append(("**", args_node.kwarg, annotation, None))
+
+    return result
+
+
+def format_args(args_node):
+    result = []
+
+    args_info = _get_args_info(args_node)
+    for prefix, name, annotation, default in args_info:
+        formatted = "{}{}{}{}".format(
+            prefix or "",
+            name or "",
+            ": {}".format(annotation) if annotation else "",
+            (" = {}" if annotation else "={}").format(default) if default else "",
+        )
+        result.append(formatted)
 
     return ", ".join(result)
 
 
+def get_annotations_dict(args_node):
+    result = collections.OrderedDict()
+
+    args_info = _get_args_info(args_node)
+    for _, name, annotation, __ in args_info:
+        if name and annotation:
+            result[name] = annotation
+
+    return result
+
+
 def get_func_docstring(node):
     """Get the docstring of a node, using a parent docstring if needed.
 
